local helpers = require "spec.helpers"
local cjson = require "cjson"

local jwt_secrets = helpers.dao.jwt_secrets
local fixtures = require "spec.03-plugins.jwt.fixtures"

describe("Plugin: jwt (API)", function()
  local admin_client, consumer, jwt_secret
  setup(function()
    helpers.prepare_prefix()
    assert(helpers.start_kong())
    admin_client = assert(helpers.admin_client())
  end)
  teardown(function()
    if admin_client then admin_client:close() end
    assert(helpers.stop_kong())
    helpers.clean_prefix()
  end)

  describe("/consumers/:consumer/jwt/", function()
    setup(function()
      consumer = assert(helpers.dao.consumers:insert {
        username = "bob"
      })
    end)

    describe("POST", function()
      local jwt1, jwt2
      teardown(function()
        if jwt1 == nil then return end
        jwt_secrets:delete(jwt1)
        jwt_secrets:delete(jwt2)
      end)

      it("creates a jwt secret", function()
        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {},
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        local body = cjson.decode(assert.res_status(201, res))
        assert.equal(consumer.id, body.consumer_id)
        jwt1 = body
      end)
      it("accepts any given `secret` and `key` parameters", function()
        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {
            key = "bob2",
            secret = "tooshort"
          },
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        local body = cjson.decode(assert.res_status(201, res))
        assert.equal("bob2", body.key)
        assert.equal("tooshort", body.secret)
        jwt2 = body
      end)
      it("accepts a valid public key for RS256 when posted urlencoded", function()
        local rsa_public_key = fixtures.rs256_public_key
        rsa_public_key = rsa_public_key:gsub("\n", "\r\n")
        rsa_public_key = rsa_public_key:gsub("([^%w %-%_%.%~])",
          function(c) return string.format ("%%%02X", string.byte(c)) end)
        rsa_public_key = rsa_public_key:gsub(" ", "+")

        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {
            key = "bob3",
            algorithm = "RS256",
            rsa_public_key = rsa_public_key
          },
          headers = {
            ["Content-Type"] = "application/x-www-form-urlencoded"
          }
        })
        assert.response(res).has.status(201)
        local json = assert.response(res).has.jsonbody()
        assert.equal("bob3", json.key)
      end)
      it("accepts a valid public key for RS256 when posted as json", function()
        local rsa_public_key = fixtures.rs256_public_key

        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {
            key = "bob4",
            algorithm = "RS256",
            rsa_public_key = rsa_public_key
          },
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        assert.response(res).has.status(201)
        local json = assert.response(res).has.jsonbody()
        assert.equal("bob4", json.key)
      end)
      it("fails with missing `rsa_public_key` parameter for RS256 algorithms", function ()
        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {
            key = "bob5",
            algorithm = "RS256"
          },
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        assert.response(res).has.status(400)
        local json = assert.response(res).has.jsonbody()
        assert.equal("no mandatory 'rsa_public_key'", json.message)
      end)
      it("fails with an invalid rsa_public_key for RS256 algorithms", function ()
        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {
            key = "bob5",
            algorithm = "RS256",
            rsa_public_key = "test",
          },
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        assert.response(res).has.status(400)
        local json = assert.response(res).has.jsonbody()
        assert.equal("'rsa_public_key' format is invalid", json.message)
      end)
      it("does not fail when `secret` parameter for HS256 algorithms is missing", function ()
        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {
            key = "bob5",
            algorithm = "HS256",
          },
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        assert.response(res).has.status(201)
        local json = assert.response(res).has.jsonbody()
        assert.string(json.secret)
        assert.equals(32, #json.secret)
        assert.matches("^%x+$", json.secret)
      end)
    end)

    describe("PUT", function()
      it("creates and update", function()
        local res = assert(admin_client:send {
          method = "POST",
          path = "/consumers/bob/jwt/",
          body = {},
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        local body = cjson.decode(assert.res_status(201, res))
        assert.equal(consumer.id, body.consumer_id)

        -- For GET tests
        jwt_secret = body
      end)
    end)

    describe("GET", function()
      it("retrieves all", function()
        local res = assert(admin_client:send {
          method = "GET",
          path = "/consumers/bob/jwt/",
        })
        local body = cjson.decode(assert.res_status(200, res))
        assert.equal(4, #(body.data))
      end)
    end)
  end)

  describe("/consumers/:consumer/jwt/:id", function()
    describe("GET", function()
      it("retrieves by id", function()
        local res = assert(admin_client:send {
          method = "GET",
          path = "/consumers/bob/jwt/"..jwt_secret.id,
        })
        assert.res_status(200, res)
      end)
    end)

    describe("PATCH", function()
      it("updates a credential", function()
        local res = assert(admin_client:send {
          method = "PATCH",
          path = "/consumers/bob/jwt/"..jwt_secret.id,
          body = {
            key = "alice",
            secret = "newsecret"
          },
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        local body = assert.res_status(200, res)
        jwt_secret = cjson.decode(body)
        assert.equal("newsecret", jwt_secret.secret)
      end)
    end)

    describe("DELETE", function()
      it("deletes a credential", function()
        local res = assert(admin_client:send {
          method = "DELETE",
          path = "/consumers/bob/jwt/"..jwt_secret.id,
          body = {},
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        assert.res_status(204, res)
      end)
      it("returns proper errors", function()
        local res = assert(admin_client:send {
          method = "DELETE",
          path = "/consumers/bob/jwt/".."blah",
          body = {},
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        assert.res_status(400, res)

       local res = assert(admin_client:send {
          method = "DELETE",
          path = "/consumers/bob/jwt/".."00000000-0000-0000-0000-000000000000",
          body = {},
          headers = {
            ["Content-Type"] = "application/json"
          }
        })
        assert.res_status(404, res)
      end)
    end)
  end)
end)
